var Util = require('./Util');
var Escaping;
(function (Escaping) {
    Escaping[Escaping["LITERAL"] = 0] = "LITERAL";
    Escaping[Escaping["NORMAL"] = 1] = "NORMAL";
    Escaping[Escaping["TITLE"] = 2] = "TITLE";
    Escaping[Escaping["URL"] = 3] = "URL";
})(Escaping || (Escaping = {}));
var MarkdownRenderer = (function () {
    function MarkdownRenderer(options) {
        this.prefix = '';
        this.neededNewLines = 0;
        this.options = Util.assign({ preserveHardbreaks: true, preserveSoftbreaks: true }, options);
    }
    MarkdownRenderer.prototype.lastChar = function () {
        return this.buffer.charAt(this.buffer.length - 1);
    };
    MarkdownRenderer.prototype.truncatePrefix = function (length) {
        this.prefix = this.prefix.substr(0, this.prefix.length - length);
    };
    MarkdownRenderer.prototype.blankline = function () {
        if (this.neededNewLines < 2) {
            this.neededNewLines = 2;
        }
    };
    MarkdownRenderer.prototype.newLine = function () {
        if (this.neededNewLines < 1) {
            this.neededNewLines = 1;
        }
    };
    MarkdownRenderer.prototype.literal = function (s) {
        this.out(s, false, Escaping.LITERAL);
    };
    MarkdownRenderer.prototype.outputCharacter = function (escape, c, nextC) {
        var needsEscaping = false;
        var encoded;
        needsEscaping = escape != Escaping.LITERAL &&
            ((escape == Escaping.NORMAL &&
                (c == '*' || c == '_' || c == '[' || c == ']' || c == '#' || c == '<' ||
                    c == '>' || c == '\\' || c == '`' || c == '!' ||
                    (c == '&' && Util.isAlpha(nextC)) || (c == '!' && nextC == '[') ||
                    (this.isBeginLine && (c == '-' || c == '+' || c == '=')) ||
                    ((c == '.' || c == ')') &&
                        Util.isDigit(this.lastChar())))) ||
                (escape == Escaping.URL && (c == '`' || c == '<' || c == '>' || Util.isSpace(c) ||
                    c == '\\' || c == ')' || c == '(')) ||
                (escape == Escaping.TITLE &&
                    (c == '`' || c == '<' || c == '>' || c == '"' || c == '\\')));
        if (needsEscaping) {
            if (Util.isSpace(c)) {
                this.buffer += '%20';
            }
            else {
                this.buffer += '\\';
                this.buffer += c;
            }
        }
        else {
            this.buffer += c;
        }
    };
    MarkdownRenderer.prototype.out = function (source, wrap, escape) {
        var nextc;
        var c;
        var i = 0;
        var len;
        var remainder = "";
        var k = this.buffer.length - 1;
        wrap = wrap && !this.isNoWrap;
        if (this.inTightListItem && this.neededNewLines > 1) {
            this.neededNewLines = 1;
        }
        while (this.neededNewLines) {
            if (k < 0 || this.buffer[k] == '\n') {
                k -= 1;
            }
            else {
                this.buffer += '\n';
                if (this.neededNewLines > 1) {
                    this.buffer += this.prefix;
                }
            }
            this.isBeginLine = true;
            this.neededNewLines -= 1;
        }
        while (i < source.length) {
            if (this.isBeginLine) {
                this.buffer += this.prefix;
            }
            c = source[i];
            nextc = source[i + 1];
            if (c === ' ' && wrap) {
                if (!this.isBeginLine) {
                    this.buffer += ' ';
                    this.isBeginLine = false;
                    while (source[i + 1] == ' ') {
                        i++;
                    }
                }
            }
            else if (c === '\n') {
                this.buffer += '\n';
                this.isBeginLine = true;
            }
            else if (escape === Escaping.LITERAL) {
                this.buffer += c;
                this.isBeginLine = false;
            }
            else {
                this.outputCharacter(escape, c, nextc);
                this.isBeginLine = false;
            }
            i++;
        }
    };
    MarkdownRenderer.prototype.shorterstUnusedBacktickSequence = function (code) {
        var used = 1;
        var current = 0;
        var i = 0;
        while (i <= code.length) {
            if (code[i] == '`') {
                current++;
            }
            else {
                if (current) {
                    used |= (1 << current);
                }
                current = 0;
            }
            i++;
        }
        i = 0;
        while (used & 1) {
            used = used >> 1;
            i++;
        }
        return i;
    };
    MarkdownRenderer.prototype.longestBacktickSequence = function (code) {
        var longest = 0;
        var current = 0;
        var i = 0;
        var code_len = code.length;
        while (i <= code_len) {
            if (code[i] == '`') {
                current++;
            }
            else {
                if (current > longest) {
                    longest = current;
                }
                current = 0;
            }
            i++;
        }
        return longest;
    };
    MarkdownRenderer.prototype.isContainer = function (node) {
        return ['Document', 'BlockQuote', 'List', 'Item', 'CodeBlock', 'HtmlBlock', 'Paragraph', 'Header', 'HorizontalRule'].indexOf(node.type) >= 0;
    };
    MarkdownRenderer.prototype.getContainingBlock = function (node) {
        while (node && !this.isContainer(node)) {
            node = node.parent;
        }
        return node;
    };
    MarkdownRenderer.prototype.isTightList = function (node) {
        if (node && node.type === 'List') {
            return node.listTight;
        }
        else {
            return false;
        }
    };
    MarkdownRenderer.prototype.consolidateTextNodes = function (root) {
        var text = '';
        var walker = root.walker();
        var current;
        walker.next();
        while ((current = walker.next()) && current.node !== root) {
            if (current.entering) {
                var currentNode = current.node;
                if (currentNode.type === 'Text' && currentNode.next && currentNode.next.type === 'Text') {
                    text += currentNode.literal;
                }
                else {
                    return false;
                }
            }
        }
        return text;
    };
    MarkdownRenderer.prototype.isAutolink = function (node) {
        var title, url, linkText, realUrl;
        url = node.destination;
        if (url.length === 0 || url.indexOf(':') > -1) {
            return false;
        }
        title = node.title;
        if (title.length > 0) {
            return false;
        }
        linkText = node.firstChild;
        var text = this.consolidateTextNodes(linkText);
        if (text !== false) {
            realUrl = url;
            if (realUrl.substr(0, 7) === 'mailto:') {
                realUrl += 7;
            }
            return realUrl === text;
        }
        else {
            return false;
        }
    };
    MarkdownRenderer.prototype.renderNode = function (node, entering) {
        if (!node._listData) {
            node._listData = {};
        }
        if (!(node.listType === 'Item' && !node.prev && entering)) {
            var tmp = this.getContainingBlock(node);
            this.inTightListItem =
                (tmp.type === 'Item' &&
                    this.isTightList(tmp.parent)) ||
                    (tmp && tmp.parent && tmp.parent.type === 'Item' &&
                        this.isTightList(tmp.parent.parent));
        }
        switch (node.type) {
            case 'Document':
                break;
            case 'BlockQuote':
                if (entering) {
                    this.literal("> ");
                    this.prefix += "> ";
                }
                else {
                    this.prefix = this.prefix.substr(0, this.prefix.length - 2);
                    this.blankline();
                }
                break;
            case 'List':
                if (!entering && node.next && (node.next.type == 'CodeBlock' ||
                    node.next.type == 'List')) {
                    this.literal("\n");
                }
                break;
            case 'Item':
                var markerWidth, listMarker;
                if (node.parent.listType === 'Bullet') {
                    markerWidth = 2;
                }
                else {
                    var listNumber = node.parent.listStart || 1;
                    var listDelimiter = node.parent.listDelimiter;
                    var tmp = node;
                    while (tmp.prev) {
                        tmp = tmp.prev;
                        listNumber += 1;
                    }
                    listMarker = listNumber + listDelimiter;
                    if (listNumber < 10) {
                        listMarker += '  ';
                    }
                    else {
                        listMarker += ' ';
                    }
                    markerWidth = listMarker.length;
                }
                if (entering) {
                    if (node.parent.listType === 'Bullet') {
                        this.literal("* ");
                        this.prefix += "  ";
                    }
                    else {
                        this.literal(listMarker);
                        for (var i = markerWidth; i > 0; i--) {
                            this.prefix += ' ';
                        }
                    }
                }
                else {
                    this.truncatePrefix(markerWidth);
                    this.newLine();
                }
                break;
            case 'Header':
                if (entering) {
                    for (var i = node.level; i > 0; i--) {
                        this.literal("#");
                    }
                    this.literal(" ");
                    this.isNoWrap = true;
                }
                else {
                    this.isNoWrap = false;
                    this.blankline();
                }
                break;
            case 'CodeBlock':
                this.blankline();
                var info = node.info || '';
                var code = node.literal;
                if (!info.length &&
                    (code.length > 2 && !Util.isSpace(code[0])) &&
                    !(Util.isSpace(code[code.length - 1]) && Util.isSpace(code[code.length - 2])) &&
                    !(!node.prev && node.parent && node.parent.type == 'Item')) {
                    this.literal("    ");
                    this.prefix += "    ";
                    this.out(code, false, Escaping.LITERAL);
                    this.truncatePrefix(4);
                }
                else {
                    var numticks_1 = this.longestBacktickSequence(code) + 1;
                    if (numticks_1 < 3) {
                        numticks_1 = 3;
                    }
                    for (var i = 0; i < numticks_1; i++) {
                        this.literal("`");
                    }
                    this.literal(" ");
                    this.out(info, false, Escaping.LITERAL);
                    this.newLine();
                    this.literal(node.literal);
                    this.newLine();
                    for (var i = 0; i < numticks_1; i++) {
                        this.literal("`");
                    }
                }
                this.blankline();
                break;
            case 'HtmlBlock':
                this.blankline();
                this.out(node.literal, false, Escaping.LITERAL);
                this.blankline();
                break;
            case 'HorizontalRule':
                this.blankline();
                this.literal("-----");
                this.blankline();
                break;
            case 'Paragraph':
                if (!entering) {
                    this.blankline();
                }
                break;
            case 'Text':
                this.out(node.literal, false, Escaping.NORMAL);
                break;
            case 'Hardbreak':
                if (this.options.preserveHardbreaks) {
                    this.literal("\\");
                }
                this.newLine();
                break;
            case 'Softbreak':
                if (this.options.preserveSoftbreaks) {
                    this.newLine();
                }
                else {
                    this.out(" ", true, Escaping.LITERAL);
                }
                break;
            case 'Code':
                code = node.literal;
                var numticks = this.shorterstUnusedBacktickSequence(code);
                for (var i = 0; i < numticks; i++) {
                    this.literal("`");
                }
                if (code.length == 0 || code[0] == '`') {
                    this.literal(" ");
                }
                this.out(code, true, Escaping.LITERAL);
                if (code.length == 0 || code[code.length - 1] == '`') {
                    this.literal(" ");
                }
                for (var i = 0; i < numticks; i++) {
                    this.literal("`");
                }
                break;
            case 'Html':
                this.out(node.literal, false, Escaping.LITERAL);
                break;
            case 'Strong':
                if (entering) {
                    this.literal("**");
                }
                else {
                    this.literal("**");
                }
                break;
            case 'Emph':
                var emphesisDelimiter;
                if (node.parent && node.parent.type == 'Emph' &&
                    !node.next && !node.prev) {
                    emphesisDelimiter = "_";
                }
                else {
                    emphesisDelimiter = "*";
                }
                if (entering) {
                    this.literal(emphesisDelimiter);
                }
                else {
                    this.literal(emphesisDelimiter);
                }
                break;
            case 'Link':
                if (this.isAutolink(node)) {
                    if (entering) {
                        this.literal("<");
                        if (node.destination.substr(0, 7) === 'mailto:') {
                            this.literal(node.destination.substr(7));
                        }
                        else {
                            this.literal(node.destination);
                        }
                        this.literal(">");
                        return false;
                    }
                }
                else {
                    if (entering) {
                        this.literal("[");
                    }
                    else {
                        this.literal("](");
                        this.out(node.destination, false, Escaping.URL);
                        var title = node.title || '';
                        if (title.length > 0) {
                            this.literal(" \"");
                            this.out(title, false, Escaping.TITLE);
                            this.literal("\"");
                        }
                        this.literal(")");
                    }
                }
                break;
            case 'Image':
                if (entering) {
                    this.literal("![");
                }
                else {
                    this.literal("](");
                    this.out(node.destination, false, Escaping.URL);
                    var title = node.title || '';
                    if (title.length > 0) {
                        this.out(" \"", true, Escaping.LITERAL);
                        this.out(title, false, Escaping.TITLE);
                        this.literal("\"");
                    }
                    this.literal(")");
                }
                break;
        }
        return true;
    };
    MarkdownRenderer.prototype.render = function (root) {
        this.buffer = '';
        var current;
        var walker = root.walker();
        var step;
        while (step = walker.next()) {
            current = step.node;
            if (!this.renderNode(current, step.entering)) {
                walker.resumeAt(current, false);
            }
        }
        if (this.lastChar() != '\n') {
            this.buffer += '\n';
        }
        return this.buffer;
    };
    return MarkdownRenderer;
})();
module.exports = MarkdownRenderer;
//# sourceMappingURL=MarkdownRenderer.js.map